"""
priors.py
=========

Utility functions for setting priors

Created by Maxim Ziatdinov (email: maxim.ziatdinov@gmail.com)
"""

import inspect

from typing import Union, Dict, Type, Callable

import numpyro
import jax.numpy as jnp


def place_normal_prior(param_name: str, loc: float = 0.0, scale: float = 1.0):
    """
    Samples a value from a normal distribution with the specified mean (loc) and standard deviation (scale),
    and assigns it to a named random variable in the probabilistic model. Can be useful for defining prior mean functions
    in structured Gaussian processes.
    """
    return numpyro.sample(param_name, normal_dist(loc, scale))


def place_lognormal_prior(param_name: str, loc: float = 0.0, scale: float = 1.0):
    """
    Samples a value from a log-normal distribution with the specified mean (loc) and standard deviation (scale),
    and assigns it to a named random variable in the probabilistic model. Can be useful for defining prior mean functions
    in structured Gaussian processes.
    """
    return numpyro.sample(param_name, lognormal_dist(loc, scale))


def place_halfnormal_prior(param_name: str, scale: float = 1.0):
    """
    Samples a value from a half-normal distribution with the specified standard deviation (scale),
    and assigns it to a named random variable in the probabilistic model. Can be useful for defining prior mean functions
    in structured Gaussian processes.
    """
    return numpyro.sample(param_name, halfnormal_dist(scale))


def place_uniform_prior(param_name: str,
                        low: float = None,
                        high: float = None,
                        X: jnp.ndarray = None):
    """
    Samples a value from a uniform distribution with the specified low and high values,
    and assigns it to a named random variable in the probabilistic model. Can be useful for defining prior mean functions
    in structured Gaussian processes.
    """
    d = uniform_dist(low, high, X)
    return numpyro.sample(param_name, d)


def place_gamma_prior(param_name: str,
                      c: float = None,
                      r: float = None,
                      X: jnp.ndarray = None):
    """
    Samples a value from a uniform distribution with the specified concentration (c) and rate (r) values,
    and assigns it to a named random variable in the probabilistic model. Can be useful for defining prior mean functions
    in structured Gaussian processes.
    """
    d = gamma_dist(c, r, X)
    return numpyro.sample(param_name, d)


def normal_dist(loc: float = None, scale: float = None
                ) -> numpyro.distributions.Distribution:
    """
    Generate a Normal distribution based on provided center (loc) and standard deviation (scale) parameters.
    If neither are provided, uses 0 and 1 by default. It can be used to pass custom priors to GP models.

    Examples:

        Assign custom prior to kernel lengthscale during GP model initialization
    
        >>> model = gpax.ExactGP(input_dim, kernel, lengthscale_prior_dist=gpax.priors.normal_dist(5, 1))
    
        Train as usual
    
        >>> model.fit(rng_key, X, y)

    """
    loc = loc if loc is not None else 0.0
    scale = scale if scale is not None else 1.0
    return numpyro.distributions.Normal(loc, scale)


def lognormal_dist(loc: float = None, scale: float = None) -> numpyro.distributions.Distribution:
    """
    Generate a LogNormal distribution based on provided center (loc) and standard deviation (scale) parameters.
    If neither are provided, uses 0 and 1 by default. It can be used to pass custom priors to GP models.

    Examples:

        Assign custom prior to kernel lengthscale during GP model initialization
    
        >>> model = gpax.ExactGP(input_dim, kernel, lengthscale_prior_dist=gpax.priors.lognormal_dist(0, 0.1))
    
        Train as usual
    
        >>> model.fit(rng_key, X, y)

    """
    loc = loc if loc is not None else 0.0
    scale = scale if scale is not None else 1.0
    return numpyro.distributions.LogNormal(loc, scale)


def halfnormal_dist(scale: float = None) -> numpyro.distributions.Distribution:
    """
    Generate a half-normal distribution based on provided standard deviation (scale).
    If none is provided, uses 1.0 by default. It can be used to pass custom priors to GP models.

    Examples:

        Assign custom prior to noise variance during GP model initialization
    
        >>> model = gpax.ExactGP(input_dim, kernel, noise_prior_dist=gpax.priors.halfnormal_dist(0.1))
    
        Train as usual
    
        >>> model.fit(rng_key, X, y)

    """
    scale = scale if scale is not None else 1.0
    return numpyro.distributions.HalfNormal(scale)


def gamma_dist(c: float = None,
               r: float = None,
               input_vec: jnp.ndarray = None
               ) -> numpyro.distributions.Distribution:
    """
    Generate a Gamma distribution based on provided shape (c) and rate (r) parameters. If the shape (c) is not provided,
    it attempts to infer it using the range of the input vector divided by 2. The rate parameter defaults to 1.0 if not provided.
    It can be used to pass custom priors to GP models.

    Examples:

        Assign custom prior to kernel lengthscale during GP model initialization
    
        >>> model = gpax.ExactGP(input_dm, kernel, lengthscale_prior_dist=gpax.priors.gamma_dist(2, 5))
    
        Train as usual
    
        >>> model.fit(rng_key, X, y)

    """
    if c is None:
        if input_vec is not None:
            c = (input_vec.max() - input_vec.min()) / 2
        else:
            raise ValueError("Provide either c or an input array")
    if r is None:
        r = 1.0
    return numpyro.distributions.Gamma(c, r)


def uniform_dist(low: float = None,
                 high: float = None,
                 input_vec: jnp.ndarray = None
                 ) -> numpyro.distributions.Distribution:
    """
    Generate a Uniform distribution based on provided low and high bounds. If one of the bounds is not provided,
    it attempts to infer the missing bound(s) using the minimum or maximum value from the input vector.
    It can be used to pass custom priors to GP models.

    Examples:

        Assign custom prior to kernel lengthscale during GP model initialization
    
        >>> model = gpax.ExactGP(input_dm, kernel, lengthscale_prior_dist=gpax.priors.uniform_dist(1, 3))
    
        Train as usual
    
        >>> model.fit(rng_key, X, y)
    """
    if (low is None or high is None) and input_vec is None:
        raise ValueError(
            "If 'low' or 'high' is not provided, an input array must be provided.")
    low = low if low is not None else input_vec.min()
    high = high if high is not None else input_vec.max()

    return numpyro.distributions.Uniform(low, high)


def auto_priors(func: Callable, params_begin_with: int, dist_type: str = 'normal', loc: float = 0.0, scale: float = 1.0) -> Callable:
    """
    Generates a function that, when invoked, samples from normal or log-normal distributions
    for each parameter of the given deterministic function, except the first one.

    Args:
        func (Callable): The deterministic function for which to set normal or log-normal priors.
        params_begin_with (int): Parameters to account for start from this number.
        loc (float, optional): Mean of the normal or log-normal distribution. Defaults to 0.0.
        scale (float, optional): Standard deviation of the normal or log-normal distribution. Defaults to 1.0.

    Returns:
        A function that, when invoked, returns a dictionary of sampled values
        from normal or log-normal distributions for each parameter of the original function.
    """
    place_prior = place_lognormal_prior if dist_type == 'lognormal' else place_normal_prior

    # Get the names of the parameters of the function excluding the first one (dependent variable)
    params_names = list(inspect.signature(func).parameters.keys())[params_begin_with:]

    def sample_priors() -> Dict[str, Union[float, Type[Callable]]]:
        # Return a dictionary with normal priors for each parameter
        return {name: place_prior(name, loc, scale) for name in params_names}

    return sample_priors


def auto_normal_priors(func: Callable, loc: float = 0.0, scale: float = 1.0) -> Callable:
    """
    Places normal priors over function parameters.

    Args:
        func (Callable): The deterministic function for which to set normal priors.
        loc (float, optional): Mean of the normal distribution. Defaults to 0.0.
        scale (float, optional): Standard deviation of the normal distribution. Defaults to 1.0.

    Returns:
        A function that, when invoked, returns a dictionary of sampled values
        from normal distributions for each parameter of the original function.
    """
    return auto_priors(func, 1, 'normal', loc, scale)


def auto_lognormal_priors(func: Callable, loc: float = 0.0, scale: float = 1.0) -> Callable:
    """
    Places log-normal priors over function parameters.

    Args:
        func (Callable): The deterministic function for which to set log-normal priors.
        loc (float, optional): Mean of the log-normal distribution. Defaults to 0.0.
        scale (float, optional): Standard deviation of the log-normal distribution. Defaults to 1.0.

    Returns:
        A function that, when invoked, returns a dictionary of sampled values
        from log-normal distributions for each parameter of the original function.
    """
    return auto_priors(func, 1, 'lognormal', loc, scale)


def auto_normal_kernel_priors(kernel_fn: Callable, loc: float = 0.0, scale: float = 1.0) -> Callable:
    """
    Places normal priors over the kernel parameters.

    Args:
        func (Callable): The deterministic kernel function for which to set normal priors.
        loc (float, optional): Mean of the normal distribution. Defaults to 0.0.
        scale (float, optional): Standard deviation of the normal distribution. Defaults to 1.0.

    Returns:
        A function that, when invoked, returns a dictionary of sampled values
        from normal distributions for each parameter of the original kernel function.
    """
    return auto_priors(kernel_fn, 2, 'normal', loc, scale)


def auto_lognormal_kernel_priors(kernel_fn: Callable, loc: float = 0.0, scale: float = 1.0) -> Callable:
    """
    Places log-normal priors over the kernel parameters.

    Args:
        func (Callable): The deterministic kernel function for which to set log-normal priors.
        loc (float, optional): Mean of the log-normal distribution. Defaults to 0.0.
        scale (float, optional): Standard deviation of the log-normal distribution. Defaults to 1.0.

    Returns:
        A function that, when invoked, returns a dictionary of sampled values
        from log-normal distributions for each parameter of the original kernel function.
    """
    return auto_priors(kernel_fn, 2, 'lognormal', loc, scale)



import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from numpyro.infer import SVI, Trace_ELBO
from numpyro.infer.autoguide import AutoDiagonalNormal
from numpyro.optim import Adam
import jax

class LinReg:
    """Simple linear regression model"""
    def __init__(self):
        self.params = None

    @staticmethod
    def model(x, y=None):
        beta = numpyro.sample('beta', dist.Normal(jnp.zeros(x.shape[1]), 10*jnp.ones(x.shape[1])))
        alpha = numpyro.sample('alpha', dist.Normal(0, 10))
        sigma = numpyro.sample('sigma', dist.HalfCauchy(1))

        mu = alpha + jnp.dot(x, beta)
        with numpyro.plate('data', x.shape[0]):
            numpyro.sample('obs', dist.Normal(mu, sigma), obs=y)

    def train(self, x, y, learning_rate=0.01, num_iterations=5000):
        guide = AutoDiagonalNormal(self.model)
        optimizer = Adam(step_size=learning_rate)
        svi = SVI(self.model, guide, optimizer, loss=Trace_ELBO(), x=x, y=y)

        params, _, _ = svi.run(jax.random.PRNGKey(0), num_iterations, progress_bar=False)
        self.params = svi.guide.median(params)

    def predict(self, x_new):
        alpha = self.params['alpha']
        beta = self.params['beta']
        sigma = self.params['sigma']
        return alpha + jnp.dot(x_new, beta)

    def get_params(self):
        return self.params



"""
hskgp.py
=========

Fully Bayesian implementation of heteroskedastic Gaussian process regression

Created by Maxim Ziatdinov (email: maxim.ziatdinov@gmail.com)
"""

from typing import Callable, Dict, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist

from . import ExactGP
from ..kernels import get_kernel
from ..utils import _set_noise_kernel_fn

kernel_fn_type = Callable[[jnp.ndarray, jnp.ndarray, Dict[str, jnp.ndarray], jnp.ndarray], jnp.ndarray]


class VarNoiseGP(ExactGP):
    """
    Heteroskedastic Gaussian process class

    Args:
        input_dim:
            Number of input dimensions
        kernel:
            Main kernel function ('RBF', 'Matern', 'Periodic', or custom function)
        noise_kernel:
            Noise kernel function ('RBF', 'Matern', 'Periodic', or custom function)
        mean_fn:
            Optional deterministic mean function (use 'mean_fn_priors' to make it probabilistic)
        kernel_prior:
            Optional custom priors over main kernel hyperparameters. Use it when passing your custom kernel.
        mean_fn_prior:
            Optional priors over mean function parameters
        noise_kernel_prior:
            Optional custom priors over noise kernel hyperparameters. Use it when passing your custom kernel.
        lengthscale_prior_dist:
            Optional custom prior distribution over main kernel lengthscale. Defaults to LogNormal(0, 1).
        noise_mean_fn:
            Optional noise mean function
        noise_mean_fn_prior:
            Optional priors over noise mean function
        noise_lengthscale_prior_dist:
            Optional custom prior distribution over noise kernel lengthscale. Defaults to LogNormal(0, 1).
     Examples:

        Use two different kernels with default priors for main and noise processes

        >>> # Get random number generator keys for training and prediction
        >>> rng_key, rng_key_predict = gpax.utils.get_keys()
        >>> # Initialize model
        >>> gp_model = gpax.VarNoiseGP(input_dim=1, kernel='RBF, noise_kernel='Matern')
        >>> # Run HMC to obtain posterior samples for the GP model parameters
        >>> gp_model.fit(rng_key, X, y)
        >>> # Make a prediction on new inputs
        >>> y_pred, y_samples = gp_model.predict(rng_key_predict, X_new)
        >>> # Get the inferred noise samples (for training data)
        >>> data_variance = gp_model.get_data_var_samples()

        Specify custom kernel lengthscale priors for main and noise kernels

        >>> lscale_prior = gpax.utils.gamma_dist(5, 1)  # equivalent to numpyro.distributions.Gamma(5, 1)
        >>> noise_lscale_prior = gpax.utils.halfnormal_dist(1)  # equivalent to numpyro.distributions.HalfNormal(1)
        >>> # Initialize model
        >>> gp_model = gpax.VarNoiseGP(
        >>>    input_dim=1, kernel='RBF, noise_kernel='Matern',
        >>>    lengthscale_prior_dist=lscale_prior, noise_lengthscale_prior_dist=noise_lscale_prior)
        >>> # Run HMC to obtain posterior samples for the GP model parameters
        >>> gp_model.fit(rng_key, X, y)
        >>> # Make a prediction on new inputs
        >>> y_pred, y_samples = gp_model.predict(rng_key_predict, X_new)
        >>> # Get the inferred noise samples (for training data)
        >>> data_variance = gp_model.get_data_var_samples()
    """
    def __init__(
        self,
        input_dim: int,
        kernel: Union[str, kernel_fn_type],
        noise_kernel: Union[str, kernel_fn_type] = 'RBF',
        mean_fn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
        kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
        mean_fn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
        noise_kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
        lengthscale_prior_dist: Optional[dist.Distribution] = None,
        noise_mean_fn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
        noise_mean_fn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
        noise_lengthscale_prior_dist: Optional[dist.Distribution] = None
    ) -> None:
        args = (input_dim, kernel, mean_fn, kernel_prior, mean_fn_prior, None, None, lengthscale_prior_dist)
        super(VarNoiseGP, self).__init__(*args)
        noise_kernel_ = get_kernel(noise_kernel)
        self.noise_kernel = _set_noise_kernel_fn(noise_kernel_) if isinstance(noise_kernel, str) else noise_kernel_

        self.noise_mean_fn = noise_mean_fn
        self.noise_mean_fn_prior = noise_mean_fn_prior
        self.noise_kernel_prior = noise_kernel_prior
        self.noise_lengthscale_prior_dist = noise_lengthscale_prior_dist

    def model(self, X: jnp.ndarray, y: jnp.ndarray = None, **kwargs: float) -> None:
        """Heteroskedastic GP probabilistic model with inputs X and targets y"""
        # Initialize mean functions at zeros
        f_loc = jnp.zeros(X.shape[0])
        noise_f_loc = jnp.zeros(X.shape[0])

        # Sample noise kernel parameters
        if self.noise_kernel_prior:
            noise_kernel_params = self.noise_kernel_prior()
        else:
            noise_kernel_params = self._sample_noise_kernel_params()
        # Add noise prior mean function (if any)
        if self.noise_mean_fn is not None:
            args = [X]
            if self.noise_mean_fn_prior is not None:
                args += [self.noise_mean_fn_prior()]
            noise_f_loc += jnp.log(self.noise_mean_fn(*args)).squeeze()
        # Compute noise kernel
        k_noise = self.noise_kernel(X, X, noise_kernel_params, 0, **kwargs)
        # Compute log variance of the data points
        points_log_var = numpyro.sample(
            "log_var",
            dist.MultivariateNormal(loc=noise_f_loc, covariance_matrix=k_noise)
        )

        # Sample main kernel parameters
        if self.kernel_prior:
            kernel_params = self.kernel_prior()
        else:
            kernel_params = self._sample_kernel_params()
        # Add mean function (if any)
        if self.mean_fn is not None:
            args = [X]
            if self.mean_fn_prior is not None:
                args += [self.mean_fn_prior()]
            f_loc += self.mean_fn(*args).squeeze()
        # Compute main kernel
        k = self.kernel(X, X, kernel_params, 0, **kwargs)
        # Sample y according to the standard Gaussian process formula. Note that instead of adding a fixed noise term to the kernel as in regular GP,
        # we exponentiate and take a diagonal of the log_var samples to get the variance at each data point and add that variance to the main kernel
        numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=f_loc, covariance_matrix=k+jnp.diag(jnp.exp(points_log_var))),
            obs=y,
        )

    def _sample_noise_kernel_params(self) -> Dict[str, jnp.ndarray]:
        """
        Sample noise kernel parameters
        """
        if self.noise_lengthscale_prior_dist is not None:
            noise_length_dist = self.noise_lengthscale_prior_dist
        else:
            noise_length_dist = dist.LogNormal(0, 1)
        noise_scale = numpyro.sample("k_noise_scale", dist.LogNormal(0, 1))
        noise_length = numpyro.sample("k_noise_length", noise_length_dist)
        return {"k_noise_length": noise_length, "k_noise_scale": noise_scale}

    def get_mvn_posterior(
        self, X_new: jnp.ndarray, params: Dict[str, jnp.ndarray], *args, **kwargs
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Returns parameters (mean and cov) of multivariate normal posterior
        for a single sample of heteroskedastic GP parameters
        """
        # Main GP part
        y_residual = self.y_train.copy()
        if self.mean_fn is not None:
            args = [self.X_train, params] if self.mean_fn_prior else [self.X_train]
            y_residual -= self.mean_fn(*args).squeeze()
        # Compute main kernel matrices for train and test data
        k_pp = self.kernel(X_new, X_new, params, 0, **kwargs)
        k_pX = self.kernel(X_new, self.X_train, params, jitter=0.0)
        k_XX = self.kernel(self.X_train, self.X_train, params, 0, **kwargs)
        # Compute the predictive covariance and mean
        K_xx_inv = jnp.linalg.inv(k_XX)
        cov = k_pp - jnp.matmul(k_pX, jnp.matmul(K_xx_inv, jnp.transpose(k_pX)))
        mean = jnp.matmul(k_pX, jnp.matmul(K_xx_inv, y_residual))
        if self.mean_fn is not None:
            args = [X_new, params] if self.mean_fn_prior else [X_new]
            mean += self.mean_fn(*args).squeeze()

        # Noise GP part
        # Compute noise kernel matrices
        k_pX_noise = self.noise_kernel(X_new, self.X_train, params, jitter=0.0)
        k_XX_noise = self.noise_kernel(self.X_train, self.X_train, params, 0, **kwargs)
        # Compute noise predictive mean
        log_var_residual = params["log_var"].copy()
        if self.noise_mean_fn is not None:
            args = [self.X_train, params] if self.noise_mean_fn_prior else [self.X_train]
            log_var_residual -= jnp.log(self.noise_mean_fn(*args)).squeeze()
        K_xx_noise_inv = jnp.linalg.inv(k_XX_noise)
        predicted_log_var = jnp.matmul(k_pX_noise, jnp.matmul(K_xx_noise_inv, log_var_residual))
        if self.noise_mean_fn is not None:
            args = [X_new, params] if self.noise_mean_fn_prior else [X_new]
            predicted_log_var += jnp.log(self.noise_mean_fn(*args)).squeeze()
        predicted_noise_variance = jnp.exp(predicted_log_var)

        # Return the main GP's predictive mean and combined (main + noise) covariance matrix
        return mean, cov + jnp.diag(predicted_noise_variance)

    def get_data_var_samples(self):
        """Returns samples with inferred (training) data variance - aka noise"""
        samples = self.mcmc.get_samples()
        log_var = samples["log_var"]
        if self.noise_mean_fn is not None:
            if self.noise_mean_fn_prior is not None:
                mean_ = jax.vmap(self.noise_mean_fn, in_axes=(None, 0))(self.X_train.squeeze(), samples)
            else:
                mean_ = self.noise_mean_fn(self.X_train.squeeze())
            log_var += jnp.log(mean_)
        return jnp.exp(log_var)

    def _print_summary(self):
        samples = self.get_samples(1)
        numpyro.diagnostics.print_summary({k: v for (k, v) in samples.items() if 'log_var' not in k})



"""
dkl.py
=======

Fully Bayesian implementation of deep kernel learning

Created by Maxim Ziatdinov (email: maxim.ziatdinov@gmail.com)
"""

from functools import partial
from typing import Callable, Dict, Optional, Tuple, Union, List

import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from jax import jit

from .gp import ExactGP


class DKL(ExactGP):
    """
    Fully Bayesian implementation of deep kernel learning

    Args:
        input_dim:
            Number of input dimensions
        z_dim:
            Latent space dimensionality (defaults to 2)
        kernel:
            Kernel function ('RBF', 'Matern', 'Periodic', or custom function)
        kernel_prior:
            Optional priors over kernel hyperparameters; uses LogNormal(0,1) by default
        nn:
            Custom neural network ('feature extractor'); uses a 3-layer MLP
            with hyperbolic tangent activations by default
        nn_prior:
            Priors over the weights and biases in 'nn'; uses normal priors by default
        latent_prior:
            Optional prior over the latent space (BNN embedding); uses none by default
        hidden_dim:
            Optional custom MLP architecture. For example [16, 8, 4] corresponds to a 3-layer
            neural network backbone containing 16, 8, and 4 neurons activated by tanh(). The latent
            layer is added autoamtically and doesn't have to be spcified here. Defaults to [64, 32].

        **kwargs:
            Optional custom prior distributions over observational noise (noise_dist_prior)
            and kernel lengthscale (lengthscale_prior_dist)


    Examples:

        DKL with image patches as inputs and a 1-d vector as targets

        >>> # Get random number generator keys for training and prediction
        >>> key1, key2 = gpax.utils.get_keys()
        >>> input data dimensions are (n, height*width*channels)
        >>> data_dim = X.shape[-1]
        >>> # Initialize DKL model with 2 latent dimensions
        >>> dkl = gpax.DKL(data_dim, z_dim=2, kernel='RBF')
        >>> # Train model by parallelizing HMC chains on a single GPU
        >>> dkl.fit(key1, X, y, num_warmup=333, num_samples=333, num_chains=3, chain_method='vectorized')
        >>> # Obtain posterior mean and samples from DKL posterior at new inputs
        >>> # using batches to avoid memory overflow
        >>> y_pred, y_samples = dkl.predict_in_batches(key2, X_new)
    """

    def __init__(self, input_dim: int, z_dim: int = 2, kernel: str = 'RBF',
                 kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 nn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
                 nn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 latent_prior: Optional[Callable[[jnp.ndarray], Dict[str, jnp.ndarray]]] = None,
                 hidden_dim: Optional[List[int]] = None, **kwargs
                 ) -> None:
        super(DKL, self).__init__(input_dim, kernel, None, kernel_prior, **kwargs)
        hdim = hidden_dim if hidden_dim is not None else [64, 32]
        self.nn = nn if nn else get_mlp(hdim)
        self.nn_prior = nn_prior if nn_prior else get_mlp_prior(input_dim, z_dim, hdim)
        self.kernel_dim = z_dim
        self.latent_prior = latent_prior

    def model(self,
              X: jnp.ndarray,
              y: jnp.ndarray = None,
              **kwargs: float
              ) -> None:
        """DKL probabilistic model"""
        jitter = kwargs.get("jitter", 1e-6)
        # BNN part
        nn_params = self.nn_prior()
        z = self.nn(X, nn_params)
        if self.latent_prior:  # Sample latent variable
            z = self.latent_prior(z)
        # Sample GP kernel parameters
        if self.kernel_prior:
            kernel_params = self.kernel_prior()
        else:
            kernel_params = self._sample_kernel_params()
        # Sample noise
        noise = self._sample_noise()
        # GP's mean function
        f_loc = jnp.zeros(z.shape[0])
        # compute kernel
        k = self.kernel(z, z, kernel_params, noise, jitter=jitter)
        # Sample y according to the standard Gaussian process formula
        numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=f_loc, covariance_matrix=k),
            obs=y,
        )

    def get_mvn_posterior(self,
                          X_new: jnp.ndarray,
                          params: Dict[str, jnp.ndarray],
                          noiseless: bool = False,
                          **kwargs: float
                          ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        noise = params["noise"]
        noise_p = noise * (1 - jnp.array(noiseless, int))
        # embed data into the latent space
        z_train = self.nn(self.X_train, params)
        z_new = self.nn(X_new, params)
        # compute kernel matrices for train and new ('test') data
        k_pp = self.kernel(z_new, z_new, params, noise_p, **kwargs)
        k_pX = self.kernel(z_new, z_train, params, jitter=0.0)
        k_XX = self.kernel(z_train, z_train, params, noise, **kwargs)
        # compute the predictive covariance and mean
        K_xx_inv = jnp.linalg.inv(k_XX)
        cov = k_pp - jnp.matmul(k_pX, jnp.matmul(K_xx_inv, jnp.transpose(k_pX)))
        mean = jnp.matmul(k_pX, jnp.matmul(K_xx_inv, self.y_train))
        return mean, cov

    @partial(jit, static_argnames='self')
    def embed(self, X_new: jnp.ndarray) -> jnp.ndarray:
        """
        Embeds data into the latent space using the inferred weights
        of the DKL's Bayesian neural network
        """
        samples = self.get_samples(chain_dim=False)
        predictive = jax.vmap(lambda params: self.nn(X_new, params))
        z = predictive(samples)
        return z

    def _print_summary(self):
        list_of_keys = ["k_scale", "k_length", "noise", "period"]
        samples = self.get_samples(1)
        numpyro.diagnostics.print_summary(
            {k: v for (k, v) in samples.items() if k in list_of_keys})


def sample_weights(name: str, in_channels: int, out_channels: int) -> jnp.ndarray:
    """Sampling weights matrix"""
    w = numpyro.sample(name=name, fn=dist.Normal(
        loc=jnp.zeros((in_channels, out_channels)),
        scale=jnp.ones((in_channels, out_channels))))
    return w


def sample_biases(name: str, channels: int) -> jnp.ndarray:
    """Sampling bias vector"""
    b = numpyro.sample(name=name, fn=dist.Cauchy(
        loc=jnp.zeros((channels)), scale=jnp.ones((channels))))
    return b


def get_mlp(architecture: List[int]) -> Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]:
    """Returns a function that represents an MLP for a given architecture."""
    def mlp(X: jnp.ndarray, params: Dict[str, jnp.ndarray]) -> jnp.ndarray:
        """MLP for a single MCMC sample of weights and biases, handling arbitrary number of layers."""
        h = X
        for i in range(len(architecture)):
            h = jnp.tanh(jnp.matmul(h, params[f"w{i}"]) + params[f"b{i}"])
        # No non-linearity after the last layer
        z = jnp.matmul(h, params[f"w{len(architecture)}"]) + params[f"b{len(architecture)}"]
        return z
    return mlp


def get_mlp_prior(input_dim: int, output_dim: int, architecture: List[int]) -> Callable[[], Dict[str, jnp.ndarray]]:
    """Priors over weights and biases for a Bayesian MLP"""
    def mlp_prior():
        params = {}
        in_channels = input_dim
        for i, out_channels in enumerate(architecture):
            params[f"w{i}"] = sample_weights(f"w{i}", in_channels, out_channels)
            params[f"b{i}"] = sample_biases(f"b{i}", out_channels)
            in_channels = out_channels
        # Output layer
        params[f"w{len(architecture)}"] = sample_weights(f"w{len(architecture)}", in_channels, output_dim)
        params[f"b{len(architecture)}"] = sample_biases(f"b{len(architecture)}", output_dim)
        return params
    return mlp_prior



"""
uigp.py
=======

Fully Bayesian implementation of Gaussian process regression with uncertain (stochastic) inputs

Created by Maxim Ziatdinov (email: maxim.ziatdinov@gmail.com)
"""

import warnings
from typing import Callable, Dict, Optional, Tuple, Union

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist

from . import ExactGP

kernel_fn_type = Callable[[jnp.ndarray, jnp.ndarray, Dict[str, jnp.ndarray], jnp.ndarray], jnp.ndarray]


class UIGP(ExactGP):
    """
    Gaussian process with uncertain inputs

    This class extends the standard Gaussian Process model to handle uncertain inputs.
    It allows for incorporating the uncertainty in input data into the GP model, providing
    a more robust prediction.

    Args:
        input_dim:
            Number of input dimensions
        kernel:
            Kernel function ('RBF', 'Matern', 'Periodic', or custom function)
        mean_fn:
            Optional deterministic mean function (use 'mean_fn_priors' to make it probabilistic)
        kernel_prior:
            Optional custom priors over kernel hyperparameters. Use it when passing your custom kernel.
        mean_fn_prior:
            Optional priors over mean function parameters
        noise_prior_dist:
            Optional custom prior distribution over the observational noise variance.
            Defaults to LogNormal(0,1).
        lengthscale_prior_dist:
            Optional custom prior distribution over kernel lengthscale. Defaults to LogNormal(0, 1).
        sigma_x_prior_dist:
            Optional custom prior for the input uncertainty (sigma_x). Defaults to HalfNormal(0.1)
            under the assumption that data is normalized to (0, 1).

    Examples:

        UIGP with custom prior over sigma_x

        >>> # Get random number generator keys for training and prediction
        >>> rng_key, rng_key_predict = gpax.utils.get_keys()
        >>> # Initialize model
        >>> gp_model = gpax.UIGP(input_dim=1, kernel='Matern', sigma_x_prior_dist=gpax.utils.halfnormal_dist(0.5))
        >>> # Run HMC to obtain posterior samples for the model parameters
        >>> gp_model.fit(rng_key, X, y, num_warmup=2000, num_samples=10000)
        >>> # Make a prediction on new inputs
        >>> y_pred, y_samples = gp_model.predict(rng_key_predict, X_new)
    """
    def __init__(self,
                 input_dim: int,
                 kernel: Union[str, kernel_fn_type],
                 mean_fn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
                 kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 mean_fn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 noise_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 noise_prior_dist: Optional[dist.Distribution] = None,
                 lengthscale_prior_dist: Optional[dist.Distribution] = None,
                 sigma_x_prior_dist: Optional[dist.Distribution] = None
                 ) -> None:
        args = (input_dim, kernel, mean_fn, kernel_prior, mean_fn_prior, noise_prior, noise_prior_dist, lengthscale_prior_dist)
        super(UIGP, self).__init__(*args)
        self.sigma_x_prior_dist = sigma_x_prior_dist

    def model(self, X: jnp.ndarray, y: jnp.ndarray = None, **kwargs: float) -> None:
        """
        Gaussian process model for uncertain (stochastic) inputs
        """
        # Initialize mean function at zeros
        f_loc = jnp.zeros(X.shape[0])

        # Sample input X
        X_prime = self._sample_x(X)

        # Sample kernel parameters
        if self.kernel_prior:
            kernel_params = self.kernel_prior()
        else:
            kernel_params = self._sample_kernel_params()
        # Sample noise
        if self.noise_prior:  # this will be removed in the future releases
            noise = self.noise_prior()
        else:
            noise = self._sample_noise()
        # Add mean function (if any)
        if self.mean_fn is not None:
            args = [X_prime]
            if self.mean_fn_prior is not None:
                args += [self.mean_fn_prior()]
            f_loc += self.mean_fn(*args).squeeze()
        # compute kernel
        k = self.kernel(X_prime, X_prime, kernel_params, noise, **kwargs)
        # sample y according to the standard Gaussian process formula
        numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=f_loc, covariance_matrix=k),
            obs=y,
        )

    def _sample_x(self, X: jnp.ndarray) -> jnp.ndarray:
        """
        Samples new input values (X_prime) based on the original inputs (X)
        and prior belief about the uncertainty in those inputs.
        """
        n_samples, n_features = X.shape
        if self.sigma_x_prior_dist is not None:
            sigma_x_dist = self.sigma_x_prior_dist
        else:
            sigma_x_dist = dist.HalfNormal(.1 * jnp.ones(n_features))
        # Sample variances independently for each feature dimension
        with numpyro.plate("feature_variance_plate", self.kernel_dim):
            sigma_x = numpyro.sample("sigma_x", sigma_x_dist)
            # Sample input data using the sampled variances
            with numpyro.plate("X_prime_plate", n_samples, dim=-2):
                X_prime = numpyro.sample("X_prime", dist.Normal(X, sigma_x))
        return X_prime

    def get_mvn_posterior(
        self, X_new: jnp.ndarray, params: Dict[str, jnp.ndarray], noiseless: bool = False, **kwargs: float
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Returns parameters (mean and cov) of multivariate normal posterior
        for a single sample of UIGP parameters
        """
        X_train_prime = params["X_prime"]
        noise = params["noise"]
        noise_p = noise * (1 - jnp.array(noiseless, int))
        y_residual = self.y_train.copy()
        if self.mean_fn is not None:
            args = [X_train_prime, params] if self.mean_fn_prior else [X_train_prime]
            y_residual -= self.mean_fn(*args).squeeze()
        # compute kernel matrices for train and test data
        k_pp = self.kernel(X_new, X_new, params, noise_p, **kwargs)
        k_pX = self.kernel(X_new, X_train_prime, params, jitter=0.0)
        k_XX = self.kernel(X_train_prime, X_train_prime, params, noise, **kwargs)
        # compute the predictive covariance and mean
        K_xx_inv = jnp.linalg.inv(k_XX)
        cov = k_pp - jnp.matmul(k_pX, jnp.matmul(K_xx_inv, jnp.transpose(k_pX)))
        mean = jnp.matmul(k_pX, jnp.matmul(K_xx_inv, y_residual))
        if self.mean_fn is not None:
            args = [X_new, params] if self.mean_fn_prior else [X_new]
            mean += self.mean_fn(*args).squeeze()
        return mean, cov

    def _predict(
        self,
        rng_key: jnp.ndarray,
        X_new: jnp.ndarray,
        params: Dict[str, jnp.ndarray],
        n: int,
        noiseless: bool = False,
        **kwargs: float
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """Prediction with a single sample of UIGP parameters"""
        # Sample X_new using the learned standard deviation
        X_new_prime = dist.Normal(X_new, params["sigma_x"]).sample(rng_key, sample_shape=(n,))
        X_new_prime = X_new_prime.mean(0)
        # Get the predictive mean and covariance
        y_mean, K = self.get_mvn_posterior(X_new_prime, params, noiseless, **kwargs)
        # draw samples from the posterior predictive for a given set of parameters
        y_sampled = dist.MultivariateNormal(y_mean, K).sample(rng_key, sample_shape=(n,))
        return y_mean, y_sampled

    def _set_data(self, X: jnp.ndarray, y: Optional[jnp.ndarray] = None) -> Union[Tuple[jnp.ndarray], jnp.ndarray]:
        X = X if X.ndim > 1 else X[:, None]
        if y is not None:
            if not (X.max() == 1 and X.min() == 0) and not self.sigma_x_prior_dist:
                warnings.warn(
                    "The default `sigma_x` prior for uncertain (stochastic) inputs assumes data is "
                    "normalized to (0, 1), which is not the case for your data. Therefore, the default prior "
                    "may not be optimal for your case. Consider passing custom prior for sigma_x, for example, "
                    "`sigma_x_prior_dist=numpyro.distributions.HalfNormal(scale)` if using NumPyro directly "
                    "or `sigma_x_prior_dist=gpax.utils.halfnormal_dist(scale)` if using a GPax wrapper",
                    UserWarning,
                )
            return X, y.squeeze()
        return X

    def _print_summary(self):
        samples = self.get_samples(1)
        numpyro.diagnostics.print_summary({k: v for (k, v) in samples.items() if 'X_prime' not in k})



"""
fn.py
=====

Utilities for setting up custom mean and kernel functions

Created by Maxim Ziatdinov (email: maxim.ziatdinov@gmail.com)
"""

import inspect
import re

from typing import List, Callable, Optional, Dict

import jax
import jax.numpy as jnp

from ..kernels.kernels import square_scaled_distance, add_jitter, _sqrt


def set_fn(func: Callable) -> Callable:
    """
    Transforms the given deterministic function to use a params dictionary
    for its parameters, excluding the first one (assumed to be the dependent variable).

    Args:
    - func (Callable): The deterministic function to be transformed.

    Returns:
    - Callable: The transformed function where parameters are accessed
                from a `params` dictionary.
    """
    # Extract parameter names excluding the first one (assumed to be the dependent variable)
    params_names = list(inspect.signature(func).parameters.keys())[1:]

    # Create the transformed function definition
    transformed_code = f"def {func.__name__}(x, params):\n"

    # Retrieve the source code of the function and indent it to be a valid function body
    source = inspect.getsource(func).split("\n", 1)[1]
    source = "    " + source.replace("\n", "\n    ")

    # Replace each parameter name with its dictionary lookup using regex
    for name in params_names:
        source = re.sub(rf'\b{name}\b', f'params["{name}"]', source)

    # Combine to get the full source
    transformed_code += source

    # Define the transformed function in the local namespace
    local_namespace = {}
    exec(transformed_code, globals(), local_namespace)

    # Return the transformed function
    return local_namespace[func.__name__]


def set_kernel_fn(func: Callable,
                  independent_vars: List[str] = ["X", "Z"],
                  jit_decorator: bool = True,
                  docstring: Optional[str] = None) -> Callable:
    """
    Transforms the given kernel function to use a params dictionary for its hyperparameters.
    The resultant function will always add jitter before returning the computed kernel.

    Args:
        func (Callable): The kernel function to be transformed.
        independent_vars (List[str], optional): List of independent variable names in the function. Defaults to ["X", "Z"].
        jit_decorator (bool, optional): @jax.jit decorator to be applied to the transformed function. Defaults to True.
        docstring (Optional[str], optional): Docstring to be added to the transformed function. Defaults to None.

    Returns:
        Callable: The transformed kernel function where hyperparameters are accessed from a `params` dictionary.
    """

    # Extract parameter names excluding the independent variables
    params_names = [k for k, v in inspect.signature(func).parameters.items() if v.default == v.empty]
    for var in independent_vars:
        params_names.remove(var)

    transformed_code = ""
    if jit_decorator:
        transformed_code += "@jit" + "\n"

    additional_args = "noise: int = 0, jitter: float = 1e-6, **kwargs"
    transformed_code += f"def {func.__name__}({', '.join(independent_vars)}, params: Dict[str, jnp.ndarray], {additional_args}):\n"

    if docstring:
        transformed_code += '    """' + docstring + '"""\n'

    source = inspect.getsource(func).split("\n", 1)[1]
    lines = source.split("\n")

    for idx, line in enumerate(lines):
        # Convert all parameter names to their dictionary lookup throughout the function body
        for name in params_names:
            lines[idx] = re.sub(rf'\b{name}\b', f'params["{name}"]', lines[idx])

    # Combine lines back and then split again by return
    modified_source = '\n'.join(lines)
    pre_return, return_statement = modified_source.split('return', 1)

    # Append custom jitter code
    custom_code = f"    {pre_return.strip()}\n    k = {return_statement.strip()}\n"
    custom_code += """
    if X.shape == Z.shape:
        k += (noise + jitter) * jnp.eye(X.shape[0])
    return k
    """

    transformed_code += custom_code

    local_namespace = {"jit": jax.jit}
    exec(transformed_code, globals(), local_namespace)

    return local_namespace[func.__name__]


def _set_noise_kernel_fn(func: Callable) -> Callable:
    """
    Modifies the GPax kernel function to append "_noise" after "k" in dictionary keys it accesses.

    Args:
        func (Callable): Original function.

    Returns:
        Callable: Modified function.
    """

    # Get the source code of the function
    source = inspect.getsource(func)

    # Split the source into decorators, definition, and body
    decorators_and_def, body = source.split("\n", 1)

    # Replace all occurrences of params["k with params["k_noise in the body
    modified_body = re.sub(r'params\["k', 'params["k_noise', body)

    # Combine decorators, definition, and modified body
    modified_source = f"{decorators_and_def}\n{modified_body}"

    # Define local namespace including the jit decorator
    local_namespace = {"jit": jax.jit}

    # Execute the modified source to redefine the function in the provided namespace
    exec(modified_source, globals(), local_namespace)

    # Return the modified function
    return local_namespace[func.__name__]



"""
sparse_gp.py
============

Variational inference implementation of sparse Gaussian process regression

Created by Maxim Ziatdinov (email: maxim.ziatdinov@gmail.com)
"""

from typing import Callable, Dict, Optional, Tuple, Type

import jax
import jaxlib
import jax.numpy as jnp
from jax.scipy.linalg import cholesky, solve_triangular

import numpyro
import numpyro.distributions as dist
from numpyro.infer import SVI, Trace_ELBO

from .vigp import viGP
from ..utils import initialize_inducing_points


class viSparseGP(viGP):
    """
    Variational inference-based sparse Gaussian process

    Args:
        input_dim:
            Number of input dimensions
        kernel:
            Kernel function ('RBF', 'Matern', 'Periodic', or custom function)
        mean_fn:
            Optional deterministic mean function (use 'mean_fn_priors' to make it probabilistic)
        kernel_prior:
            Optional custom priors over kernel hyperparameters; uses LogNormal(0,1) by default
        mean_fn_prior:
            Optional priors over mean function parameters
        noise_prior_dist:
            Optional custom prior distribution over the observational noise variance.
            Defaults to LogNormal(0,1).
        lengthscale_prior_dist:
            Optional custom prior distribution over kernel lengthscale.
            Defaults to LogNormal(0, 1).
        guide:
            Auto-guide option, use 'delta' (default) or 'normal'
    """
    def __init__(self, input_dim: int, kernel: str,
                 mean_fn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
                 kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 mean_fn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 noise_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 noise_prior_dist: Optional[dist.Distribution] = None,
                 lengthscale_prior_dist: Optional[dist.Distribution] = None,
                 guide: str = 'delta') -> None:
        args = (input_dim, kernel, mean_fn, kernel_prior, mean_fn_prior, noise_prior,
                noise_prior_dist, lengthscale_prior_dist, guide)
        super(viSparseGP, self).__init__(*args)
        self.Xu = None

    def model(self,
              X: jnp.ndarray,
              y: jnp.ndarray = None,
              Xu: jnp.ndarray = None,
              **kwargs: float) -> None:
        """
        Probabilistic sparse Gaussian process regression model
        """
        if Xu is not None:
            Xu = numpyro.param("Xu", Xu)
        # Initialize mean function at zeros
        f_loc = jnp.zeros(X.shape[0])
        # Sample kernel parameters
        if self.kernel_prior:
            kernel_params = self.kernel_prior()
        else:
            kernel_params = self._sample_kernel_params()
        # Sample noise
        if self.noise_prior:  # this will be removed in the future releases
            noise = self.noise_prior()
        else:
            noise = self._sample_noise()
        D = jnp.broadcast_to(noise, (X.shape[0],) )
        # Add mean function (if any)
        if self.mean_fn is not None:
            args = [X]
            if self.mean_fn_prior is not None:
                args += [self.mean_fn_prior()]
            f_loc += self.mean_fn(*args).squeeze()
        # Compute kernel between inducing points
        Kuu = self.kernel(Xu, Xu, kernel_params, **kwargs)
        # Cholesky decomposition
        Luu = cholesky(Kuu).T
        # Compute kernel between inducing and training points
        Kuf = self.kernel(Xu, X, kernel_params)
        # Solve triangular system
        W = solve_triangular(Luu, Kuf, lower=True).T
        # Diagonal of the kernel matrix
        Kffdiag = jnp.diag(self.kernel(X, X, kernel_params, jitter=0))
        # Sum of squares computation
        Qffdiag = jnp.square(W).sum(axis=-1)
        # Trace term computation
        trace_term = (Kffdiag - Qffdiag).sum() / noise
        # Clamping the trace term
        trace_term = jnp.clip(trace_term, a_min=0)

        # VFE approximation
        numpyro.factor("trace_term", -trace_term / 2.0)

        numpyro.sample(
            "y",
            dist.LowRankMultivariateNormal(loc=f_loc, cov_factor=W, cov_diag=D),
            obs=y)

    def fit(self,
            rng_key: jnp.array, X: jnp.ndarray, y: jnp.ndarray,
            inducing_points_ratio: float = 0.1, inducing_points_selection: str = 'random',
            num_steps: int = 1000, step_size: float = 5e-3,
            progress_bar: bool = True, print_summary: bool = True,
            device: Type[jaxlib.xla_client.Device] = None,
            **kwargs: float
            ) -> None:
        """
        Run variational inference to learn sparse GP (hyper)parameters

        Args:
            rng_key: random number generator key
            X: 2D feature vector with *(number of points, number of features)* dimensions
            y: 1D target vector with *(n,)* dimensions
            Xu: Inducing points ratio. Must be a float between 0 and 1. Default value is 0.1.
            num_steps: number of SVI steps
            step_size: step size schedule for Adam optimizer
            progress_bar: show progress bar
            print_summary: print summary at the end of training
            device:
                optionally specify a cpu or gpu device on which to run the inference;
                e.g., ``device=jax.devices("cpu")[0]``
            **jitter:
                Small positive term added to the diagonal part of a covariance
                matrix for numerical stability (Default: 1e-6)
        """
        X, y = self._set_data(X, y)
        if device:
            X = jax.device_put(X, device)
            y = jax.device_put(y, device)
        Xu = initialize_inducing_points(
            X.copy(), inducing_points_ratio,
            inducing_points_selection, rng_key)
        self.X_train = X
        self.y_train = y

        optim = numpyro.optim.Adam(step_size=step_size, b1=0.5)
        self.svi = SVI(
            self.model,
            guide=self.guide_type(self.model),
            optim=optim,
            loss=Trace_ELBO(),
            X=X,
            y=y,
            Xu=Xu,
            **kwargs
        )

        self.kernel_params = self.svi.run(
            rng_key, num_steps, progress_bar=progress_bar)[0]

        self.Xu = self.kernel_params['Xu']

        if print_summary:
            self._print_summary()

    def get_mvn_posterior(self, X_new: jnp.ndarray,
                          params: Dict[str, jnp.ndarray],
                          noiseless: bool = False,
                          **kwargs: float
                          ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Returns parameters (mean and cov) of multivariate normal posterior
        for a single sample of GP parameters
        """
        noise = params["noise"]
        N = self.X_train.shape[0]
        D = jnp.broadcast_to(noise, (N,))
        noise_p = noise * (1 - jnp.array(noiseless, int))

        y_residual = self.y_train.copy()
        if self.mean_fn is not None:
            args = [self.X_train, params] if self.mean_fn_prior else [self.X_train]
            y_residual -= self.mean_fn(*args).squeeze()

        # Compute self- and cross-covariance matrices
        Kuu = self.kernel(self.Xu, self.Xu, params, **kwargs)
        Luu = cholesky(Kuu, lower=True)
        Kuf = self.kernel(self.Xu, self.X_train, params, jitter=0)

        W = solve_triangular(Luu, Kuf, lower=True)
        W_Dinv = W / D
        K = W_Dinv @ W.T
        K = K.at[jnp.diag_indices(K.shape[0])].add(1)
        L = cholesky(K, lower=True)

        y_2D = y_residual.reshape(-1, N).T
        W_Dinv_y = W_Dinv @ y_2D

        Kus = self.kernel(self.Xu, X_new, params, jitter=0)
        Ws = solve_triangular(Luu, Kus, lower=True)
        pack = jnp.concatenate((W_Dinv_y, Ws), axis=1)
        Linv_pack = solve_triangular(L, pack, lower=True)

        Linv_W_Dinv_y = Linv_pack[:, :W_Dinv_y.shape[1]]
        Linv_Ws = Linv_pack[:, W_Dinv_y.shape[1]:]
        mean = (Linv_W_Dinv_y.T @ Linv_Ws).squeeze()

        Kss = self.kernel(X_new, X_new, params, noise_p, **kwargs)
        Qss = Ws.T @ Ws
        cov = Kss - Qss + Linv_Ws.T @ Linv_Ws

        if self.mean_fn is not None:
            args = [X_new, params] if self.mean_fn_prior else [X_new]
            mean += self.mean_fn(*args).squeeze()

        return mean, cov


"""
optimize.py
==============

Optimize continuous acquisition functions

Created by Maxim Ziatdinov (email: maxim.ziatdinov@gmail.com)
"""

from typing import Type, Callable, Union, List, Tuple

import jax.numpy as jnp
import jax.random as jra
import numpy as onp

from ..models.gp import ExactGP


def optimize_acq(rng_key: jnp.ndarray,
                 model: Type[ExactGP],
                 acq_fn: Callable,
                 num_initial_guesses: int,
                 lower_bound: Union[List, Tuple, float, onp.ndarray, jnp.ndarray],
                 upper_bound: Union[List, Tuple, float, onp.ndarray, jnp.ndarray],
                 **kwargs) -> jnp.ndarray:
    """
    Optimizes an acquisition function for a given Gaussian Process model using the JAXopt library.

    This function finds the point that maximizes the acquisition function within the specified bounds.
    It uses L-BFGS-B algorithm through ScipyBoundedMinimize from JAXopt.

    Args:
        rng_key: A JAX random key for stochastic processes.
        model: The Gaussian Process model to be used.
        acq_fn: The acquisition function to be maximized.
        num_initial_guesses: Number of random initial guesses for the optimization.
        lower_bound: Lower bounds for the optimization.
        upper_bound: Upper bounds for the optimization.
        **kwargs: Additional keyword arguments to be passed to the acquisition function.

    Returns:
        Parameter(s) that maximize the acquisition function within the specified bounds.

    Note:
        Ensure JAXopt is installed to use this function (`pip install jaxopt`).
        The acquisition function is minimized using its negative value to find the maximum.

    Examples:

        Optimize EI given a trained GP model for 1D problem

        >>> acq_fn = gpax.acquisition.EI
        >>> num_initial_guesses = 10
        >>> lower_bound = -2.0
        >>> upper_bound = 2.0
        >>> x_next = gpax.acquisition.optimize_acq(
        >>>    rng_key, gp_model, acq_fn,
        >>>    num_initial_guesses, lower_bound, upper_bound,
        >>>    maximize=False, noiseless=True)
    """

    try:
        import jaxopt  # noqa: F401
    except ImportError as e:
        raise ImportError(
            "You need to install `jaxopt` to be able to use this feature. "
            "It can be installed with `pip install jaxopt`."
        ) from e

    def acq(x):
        x = jnp.array([x])
        x = x[None] if x.ndim == 0 else x
        obj = -acq_fn(rng_key, model, x, **kwargs)
        return jnp.reshape(obj, ())

    lower_bound = ensure_array(lower_bound)
    upper_bound = ensure_array(upper_bound)

    initial_guesses = jra.uniform(
        rng_key, shape=(num_initial_guesses, lower_bound.shape[0]),
        minval=lower_bound, maxval=upper_bound)
    initial_acq_vals = acq_fn(rng_key, model, initial_guesses, **kwargs)
    best_initial_guess = initial_guesses[initial_acq_vals.argmax()].squeeze()

    minimizer = jaxopt.ScipyBoundedMinimize(fun=acq, method='l-bfgs-b')
    result = minimizer.run(best_initial_guess, bounds=(lower_bound, upper_bound))

    return result.params


def ensure_array(x):
    if not isinstance(x, jnp.ndarray):
        if isinstance(x, (list, tuple, float, onp.ndarray)):
            x = jnp.array([x]) if isinstance(x, float) else jnp.array(x)
        else:
            raise TypeError(f"Expected input to be a list, tuple, float, or jnp.ndarray, got {type(x)} instead.")
    return x



from . import acquisition, kernels, priors, utils
from ._version import __version__
from .hypo import sample_next
from .models import (
    BNN,
    DKL,
    UIGP,
    CoregGP,
    ExactGP,
    MeasuredNoiseGP,
    MultiTaskGP,
    VarNoiseGP,
    iBNN,
    sPM,
    vExactGP,
    vi_iBNN,
    viDKL,
    viGP,
    viMTDKL,
    viSparseGP,
)

__all__ = [
    "priors",
    "utils",
    "kernels",
    "mtkernels",
    "acquisition",
    "ExactGP",
    "vExactGP",
    "DKL",
    "viDKL",
    "iBNN",
    "vi_iBNN",
    "MultiTaskGP",
    "viMTDKL",
    "viGP",
    "sPM",
    "VarNoiseGP",
    "UIGP",
    "MeasuredNoiseGP",
    "viSparseGP",
    "CoregGP",
    "BNN",
    "sample_next",
    "__version__",
]



"""
bnn.py
=======

Fully Bayesian MLPs

Created by Maxim Ziatdinov (email: maxim.ziatdinov@gmail.com)
"""

from typing import Callable, Dict, Optional, List, Union, Tuple

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist

from .spm import sPM


class BNN(sPM):
    """Fully Bayesian MLP"""
    def __init__(self,
                 input_dim: int,
                 output_dim: int,
                 noise_prior_dist: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 hidden_dim: Optional[List[int]] = None, **kwargs):
        hidden_dim = [64, 32] if not hidden_dim else hidden_dim
        nn = kwargs.get("nn", get_mlp(hidden_dim))
        nn_prior = kwargs.get("nn_prior", get_mlp_prior(input_dim, output_dim, hidden_dim))
        super(BNN, self).__init__(nn, nn_prior, None, noise_prior_dist)

    def _set_data(self, X: jnp.ndarray, y: Optional[jnp.ndarray] = None
                  ) -> Union[Tuple[jnp.ndarray], jnp.ndarray]:
        X = X if X.ndim > 1 else X[:, None]
        if y is not None:
            y = y[:, None] if y.ndim < 2 else y
            return X, y
        return X


def sample_weights(name: str, in_channels: int, out_channels: int) -> jnp.ndarray:
    """Sampling weights matrix"""
    w = numpyro.sample(name=name, fn=dist.Normal(
        loc=jnp.zeros((in_channels, out_channels)),
        scale=jnp.ones((in_channels, out_channels))))
    return w


def sample_biases(name: str, channels: int) -> jnp.ndarray:
    """Sampling bias vector"""
    b = numpyro.sample(name=name, fn=dist.Cauchy(
        loc=jnp.zeros((channels)), scale=jnp.ones((channels))))
    return b


def get_mlp(architecture: List[int]) -> Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]:
    """Returns a function that represents an MLP for a given architecture."""
    def mlp(X: jnp.ndarray, params: Dict[str, jnp.ndarray]) -> jnp.ndarray:
        """MLP for a single MCMC sample of weights and biases, handling arbitrary number of layers."""
        h = X
        for i in range(len(architecture)):
            h = jnp.tanh(jnp.matmul(h, params[f"w{i}"]) + params[f"b{i}"])
        # No non-linearity after the last layer
        z = jnp.matmul(h, params[f"w{len(architecture)}"]) + params[f"b{len(architecture)}"]
        return z
    return mlp


def get_mlp_prior(input_dim: int, output_dim: int, architecture: List[int]) -> Callable[[], Dict[str, jnp.ndarray]]:
    """Priors over weights and biases for a Bayesian MLP"""
    def mlp_prior():
        params = {}
        in_channels = input_dim
        for i, out_channels in enumerate(architecture):
            params[f"w{i}"] = sample_weights(f"w{i}", in_channels, out_channels)
            params[f"b{i}"] = sample_biases(f"b{i}", out_channels)
            in_channels = out_channels
        # Output layer
        params[f"w{len(architecture)}"] = sample_weights(f"w{len(architecture)}", in_channels, output_dim)
        params[f"b{len(architecture)}"] = sample_biases(f"b{len(architecture)}", output_dim)
        return params
    return mlp_prior
